@@ -165,17 +165,6 @@ class _FieldsOrder(dict):
         return fields
 
 
-def counter(items):
-    """
-    Simplest required implementation of collections.Counter. Required as 2.6
-    does not have Counter in collections.
-    """
-    results = {}
-    for item in items:
-        results[item] = results.get(item, 0) + 1
-    return results
-
-
 def iter_child_nodes(node, omit=None, _fields_order=_FieldsOrder()):
     """
     Yield all direct child nodes of *node*, that is, all fields that
@@ -537,7 +526,10 @@ class Scope(dict):
 
 
 class ClassScope(Scope):
-    pass
+    def __init__(self):
+        super().__init__()
+        # {name: node}
+        self.indirect_assignments = {}
 
 
 class FunctionScope(Scope):
@@ -548,13 +540,14 @@ class FunctionScope(Scope):
     """
     usesLocals = False
     alwaysUsed = {'__tracebackhide__', '__traceback_info__',
-                  '__traceback_supplement__'}
+                  '__traceback_supplement__', '__debuggerskip__'}
 
     def __init__(self):
         super().__init__()
         # Simplify: manage the special locals as globals
         self.globals = self.alwaysUsed.copy()
-        self.returnValue = None     # First non-empty return
+        # {name: node}
+        self.indirect_assignments = {}
 
     def unused_assignments(self):
         """
@@ -842,6 +835,10 @@ class Checker:
         which were imported but unused.
         """
         for scope in self.deadScopes:
+            if isinstance(scope, (ClassScope, FunctionScope)):
+                for name, node in scope.indirect_assignments.items():
+                    self.report(messages.UnusedIndirectAssignment, node, name)
+
             # imports in classes are public members
             if isinstance(scope, ClassScope):
                 continue
@@ -993,6 +990,9 @@ class Checker:
                     self.report(messages.RedefinedWhileUnused,
                                 node, value.name, existing.source)
 
+                if isinstance(scope, (ClassScope, FunctionScope)):
+                    scope.indirect_assignments.pop(value.name, None)
+
             elif isinstance(existing, Importation) and value.redefines(existing):
                 existing.redefined.append(node)
 
@@ -1003,14 +1003,21 @@ class Checker:
         # don't treat annotations as assignments if there is an existing value
         # in scope
         if value.name not in self.scope or not isinstance(value, Annotation):
-            cur_scope_pos = -1
-            # As per PEP 572, use scope in which outermost generator is defined
-            while (
-                isinstance(value, NamedExprAssignment) and
-                isinstance(self.scopeStack[cur_scope_pos], GeneratorScope)
-            ):
-                cur_scope_pos -= 1
-            self.scopeStack[cur_scope_pos][value.name] = value
+            if isinstance(value, NamedExprAssignment):
+                # PEP 572: use scope in which outermost generator is defined
+                scope = next(
+                    scope
+                    for scope in reversed(self.scopeStack)
+                    if not isinstance(scope, GeneratorScope)
+                )
+                if value.name in scope and isinstance(scope[value.name], Annotation):
+                    # re-assignment to name that was previously only an annotation
+                    scope[value.name] = value
+                else:
+                    # it may be a re-assignment to an already existing name
+                    scope.setdefault(value.name, value)
+            else:
+                self.scope[value.name] = value
 
     def _unknown_handler(self, node):
         # this environment variable configures whether to error on unknown
@@ -1186,6 +1193,9 @@ class Checker:
             # be executed.
             return
 
+        if isinstance(self.scope, (ClassScope, FunctionScope)):
+            self.scope.indirect_assignments.pop(name, None)
+
         if isinstance(self.scope, FunctionScope) and name in self.scope.globals:
             self.scope.globals.remove(name)
         else:
@@ -1774,7 +1784,7 @@ class Checker:
             convert_to_value(key) for key in node.keys
         ]
 
-        key_counts = counter(keys)
+        key_counts = collections.Counter(keys)
         duplicate_keys = [
             key for key, count in key_counts.items()
             if count > 1
@@ -1783,7 +1793,7 @@ class Checker:
         for key in duplicate_keys:
             key_indices = [i for i, i_key in enumerate(keys) if i_key == key]
 
-            values = counter(
+            values = collections.Counter(
                 convert_to_value(node.values[index])
                 for index in key_indices
             )
@@ -1844,6 +1854,8 @@ class Checker:
                 for scope in self.scopeStack[global_scope_index + 1:]:
                     scope[node_name] = node_value
 
+                self.scope.indirect_assignments[node_name] = node
+
     NONLOCAL = GLOBAL
 
     def GENERATOREXP(self, node):
@@ -1896,12 +1908,6 @@ class Checker:
             self.report(messages.ReturnOutsideFunction, node)
             return
 
-        if (
-            node.value and
-            hasattr(self.scope, 'returnValue') and
-            not self.scope.returnValue
-        ):
-            self.scope.returnValue = node.value
         self.handleNode(node.value, node)
 
     def YIELD(self, node):
